from itertools import chain
from django.conf import settings
from django.apps import apps
from django.core.urlresolvers import resolve
from django.utils.text import capfirst

from ..utils.functools import factual, pick, omit
from ..utils.translation import get_string, get_model_names
from ..utils.urls import exists_import
from .library import register, Component, ComplexComponent, Tag
from .markup import Link, Caption


class NavigationMixin:
    def check_active(self, *args):
        return resolve(self.request.path).url_name in args


class MenuItem(NavigationMixin, Tag):
    el = 'li'

    def get_defaults(self):
        return {
            'divider': False,
            'include': True,
            'disabled': False,
            'url': None,
            'reverse': True}

    def get_nodelist(self):
        if self.props.divider or not self.props.include:
            return ''
        return self.inject(
            Link, pick(self.props, 'url', 'reverse'), self.props.caption)

    def resolve_props(self):
        result = {}
        if self.props.divider:
            return {}
        elif 'model' in self.props:
            model = apps.get_model(*self.props.model.split('.'))
            self.props.caption['title'] = get_model_names(model)[-1]
            (prefix, url) = map(get_string, get_model_names(model))

            result['url'] = url
            result['active'] = self.check_active(
                url, *map(
                    lambda suffix: '-'.join([prefix, suffix]),
                    ['add', 'edit']))
        elif not 'active' in self.props:
            if self.props.url and self.props.reverse:
                result['active'] = self.check_active(self.props.url)
            else:
                result['active'] = False
        if self.props.include is None:
            result['include'] = result.get('active', self.props.active)
        if self.props.disabled:
            result.update({'url': '#', 'reverse': False})
        result['caption'] = self.inject(Caption, self.props.caption)
        return result

    def tweak(self):
        if not self.props.include:
            self.dom.empty()
        elif self.props.divider:
            self.dom.add_class('nav-divider')
        else:
            self.dom.add_class(self.get_class_set('disabled', 'active'))
        super().tweak()


class Menu(NavigationMixin, Component):
    def get_items(self):
        return self.props.items

    def get_defaults(self):
        return {'pills': False, 'tabs': False, 'stacked': False}

    def render_item(self, item):
        if isinstance(item, list):
            return self.inject(
                DropdownMenu, {'toggle': item[0], 'menu': item[1:]})
        return self.inject(MenuItem, item)

    def DOM(self):
        return self.inject(
            Tag,
            {'el': 'ul', 'class': self.get_class_set(
                'pills', 'tabs', 'stacked', prefix='nav', include='nav')},
            ''.join(map(self.render_item, self.get_items())))


class DropdownMenu(NavigationMixin, Tag):
    el = 'li'

    def resolve_attrs(self):
        return {'class': 'dropdown'}

    def get_nodelist(self):
        return ''.join([
            self.inject(
                Link, {
                    'url': self.props.toggle.get('url', '#'),
                    'reverse': False,
                    'attrs': {
                        'class': 'dropdown-toggle',
                        'data-toggle': 'dropdown'}},
                ''.join([
                    self.inject(Caption, self.props.toggle['caption']),
                    '' if not self.props.menu else self.inject(
                        Tag, {'el': 'span', 'class': 'caret'})])),
            '' if not self.props.menu else self.inject(
                Tag,
                {'el': 'ul', 'class': 'dropdown-menu'},
                ''.join(map(
                    lambda i: self.inject(MenuItem, i),
                    self.props.menu)))])


@register.era
class MainMenu(Menu):
    def get_app_menu_items(self, cls):
        result = cls()
        result.request = self.request
        result.context = self.context
        return result.get_items()

    def get_items(self):
        return list(chain(*map(
            self.get_app_menu_items,
            factual(map(
                lambda module: getattr(
                    module,
                    ''.join([capfirst(module.__package__), 'Menu']),
                    None),
                factual(map(
                    lambda app: exists_import('.'.join([app, 'components'])),
                    settings.MODULES)))))))



from itertools import chain
from django.core.urlresolvers import reverse
from django.utils.text import capfirst

from ..utils.functools import just, call, factual, reduce_dict, omit, pick, truthful
from .library import register, Import, Component, ComplexComponent, Tag


@register.era
class Icon(Component):
    def DOM(self):
        return self.inject(
            Tag,
            {'el': 'i', 'class': ' '.join(chain(
                ['fa', 'fa-' + self.props.name],
                map(
                    lambda z: 'fa-' + z[0].format(self.props.pop(z[1])),
                    filter(
                        lambda z: z[1] in self.props,
                        zip(
                            ('{0}x', 'li', 'lg', 'fw', 'spin', 'rotate-{0}'),
                            ('size', 'list', 'large', 'fixed', 'spin', 'rotate'))))))})


class MarkedList(Tag):
    el = 'ul'

    def get_items(self):
        return self.props.items

    def get_nodelist(self):
        return ''.join(map(
            lambda item: '<li>{0}</li>'.format(item),
            self.get_items()))


class IconicList(Component):
    def get_defaults(self):
        return {'icon': 'asterisk', 'heading': 'h4'}

    def DOM(self):
        return ''.join([
            '' if not self.props.get('title') else self.inject(
                Tag,
                {'el': self.props.heading},
                capfirst(self.props.title)),
            self.inject(
                Tag,
                {'el': 'ul', 'class': 'fa-ul'},
                ''.join(map(
                    lambda content: self.inject(
                        Tag,
                        {'el': 'li'},
                        ''.join([content, self.inject(
                            Icon, {'name': self.props.icon, 'list': True})])),
                    map(str, self.get_items()))))])


@register.era
class Break(Component):
    def get_defaults(self):
        return {'x': 1, 'ruler': False}

    def resolve_props(self):
        return {
            'x': int(self.props.x),
            'tag': '<{0}r />'.format(self.props.ruler and 'h' or 'b')}

    def DOM(self):
        if 'join' in self.props:
            return self.props.tag.join(self.props.join)
        return self.props.x * self.props.tag


@register.era
class ProgressBar(Component):
    def get_defaults(self):
        return {'level': 'success'}

    def resolve_props(self):
        return {'value': str(int(self.props.value))}

    def DOM(self):
        return self.inject(
            Tag, {'class': 'progress'}, self.inject(
                Tag,
                {'attrs': {
                    'class': 'progress-bar progress-bar-' + self.props.level,
                    'style': 'width: {0}%;'.format(self.props.value)}},
                self.props.value + '%'))


@register.era
class Row(Tag):
    def resolve_attrs(self):
        return {'class': 'row'}


@register.era
class Container(Tag):
    def get_defaults(self):
        return {'fluid': False}

    def resolve_attrs(self):
        return {'class': 'container' + (self.props.fluid and '-fluid' or '')}


@register.era
class Column(Tag):
    def get_defaults(self):
        return {'constrict': False}

    def resolve_attrs(self):
        return {'class': ' '.join(reduce_dict(
            lambda k, v: 'col-{0}-{1}'.format(k, v),
            dict({'md': 12} if not self.props.constrict else {
                'md': 12 - self.props.constrict * 2,
                'md-offset': self.props.constrict
            }, **omit(self.props, 'nodelist', 'class', 'constrict'))))}


@register.era
class Link(Tag):
    el = 'a'

    def get_defaults(self):
        return {
            'newtab': False,
            'reverse': True,
            'url': 'index',
            'args': [],
            'kwargs': {},
            'rel': None,
            'qs': None}

    def resolve_props(self):
        if self.props.rel:
            return {
                'reverse': False,
                'url': '/'.join([
                    self.request.path.rstrip('/'),
                    self.props.rel])}
        return {}

    def get_url(self):
        if self.props.qs:
            return '?'.join([self.props.url, self.props.qs])
        return self.props.url

    def get_nodelist(self):
        return self.props.get('nodelist', self.get_url())

    def resolve_attrs(self):
        return {
            'target': self.props.newtab and '_blank' or '_self',
            'href': (self.props.reverse and reverse or just)(
                self.get_url(), **pick(self.props, 'args', 'kwargs'))}


@register.era
class Button(Tag):
    el = 'button'

    def get_defaults(self):
        return {'level': 'primary', 'type': 'button', 'name': 'button'}

    def resolve_attrs(self):
        return dict(
            {'class': 'btn btn-' + self.props.level},
            **pick(self.props, 'type', 'onclick'))


@register.era
class Label(Tag):
    el = 'span'

    def get_defaults(self):
        return {'level': 'default'}

    def resolve_attrs(self):
        return {'class': 'label label-' + self.props.level}


@register.era
class Alert(Tag):
    def get_defaults(self):
        return {'level': 'primary', 'dismiss': True}

    def resolve_attrs(self):
        return {'class': 'alert alert-' + self.props.level}

    def tweak(self):
        super().tweak()
        if self.props.dismiss:
            self.dom.prepend(
                self.inject(
                    Tag,
                    {'el': 'button', 'class': 'close', 'attrs': 
                        {'type': 'button', 'data-dismiss': 'alert'}},
                    self.inject(Icon, {'name': 'remove'})))


@register.era
class Well(Tag):
    def resolve_attrs(self):
        return {
            'class': ' '.join(factual([
                'well', 
                '' if not 'size' in self.props else 'well-' + self.props.size]))}


@register.era
class Caption(Component):
    def DOM(self):
        return self.inject(
            Tag, {'el': 'span', 'class': 'caption'},
                self.props.title if not 'icon' in self.props else ''.join([
                    self.inject(Icon, {'name': self.props.icon}),
                    self.inject(Tag, {'el': 'span'}, self.props.title)]))


@register.era
class Panel(ComplexComponent):
    parts = ['title', 'body']

    def get_defaults(self):
        return {'level': 'default', 'title': False}

    def DOM(self):
        return self.inject(
            Tag,
            {'class': 'panel panel-' + self.props.level},
            ''.join([
                '' if not self.props.title else self.inject(
                    Tag, {'class': 'panel-heading'}, self.inject(
                        Tag, {'name': 'h3'}, self.props.title)),
                self.inject(
                    Tag, {'class': 'panel-body'}, self.props.body)]))


@register.era
class Navbar(ComplexComponent):
    parts = ['head', 'brand', 'collapse']

    def get_defaults(self):
        return {'id': 'default', 'container': False}

    def DOM(self):
        nodelist = ''.join([
            self.inject(
                Tag,
                {'class': 'navbar-header'},
                ''.join([
                    self.inject(
                        Tag,
                        {'el': 'button', 'attrs': {
                            'class': 'navbar-toggle collapsed',
                            'data-toggle': 'collapse',
                            'data-target': '#' + self.props.id}},
                        self.inject(Icon, {'name': 'align-justify'})),
                    self.props.head,
                    '' if not self.props.brand else self.inject(
                        Link,
                        {'url': '/', 'reverse': False, 'class': 'navbar-brand'},
                        self.props.brand)])),
            self.inject(
                Tag,
                {'attrs': {'id': self.props.id, 'class': 'navbar-collapse collapse'}},
                self.props.collapse)])

        return self.inject(
            Tag,
            {'el': 'nav', 'class': 'navbar navbar-default'},
            nodelist if not self.props.container else self.inject(
                Container, {'fluid': self.props.container == 'fluid'}, nodelist))


class Table(Component):
    def slice(self, seq):
        return list(seq)[slice(*call(
            getattr(self, 'get_slice', lambda: (None, None))))]

    def get_thead_items(self):
        return self.props.get('thead', [])

    def get_tbody_items(self):
        return self.props.get('tbody', [])

    def get_defaults(self):
        return {
            'striped': False,
            'bordered': False,
            'hover': True,
            'condensed': True,
            'responsive': True}

    def render_content(self, content, cell):
        if isinstance(content, bool):
            return self.inject(
                Icon,
                {'name': (content and 'check' or 'minus') + '-circle'})
        return content

    def render_items(self, items, cell='td'):
        return ''.join(map(
            lambda row: self.inject(
                Tag,
                truthful({'el': 'tr', 'class': row.get('level')}),
                ''.join(map(
                    lambda c: self.inject(
                        Tag,
                        {'el': cell},
                        self.render_content(c, cell)),
                    self.slice(row['items'])))),
            items))

    def DOM(self):
        return self.inject(
            Tag,
            {'el': 'table', 'class': self.get_class_set(
                'striped',
                'bordered',
                'hover',
                'condensed',
                'responsive',
                prefix='table',
                include='table')},
            ''.join([
                    self.inject(
                        Tag,
                        {'el': 'thead'},
                        self.render_items(
                            [{'items': self.get_thead_items()}],
                            cell='th')),
                    self.inject(
                        Tag,
                        {'el': 'tbody'},
                        self.render_items(self.get_tbody_items()))]))



import operator
from functools import reduce
from itertools import chain
from urllib.parse import urlencode

from django.conf import settings
from django.db.models import Count, Q
from django.db.models.fields import BooleanField, FieldDoesNotExist
from django.db.models.fields.related import ManyToOneRel
from django.shortcuts import redirect
from django.utils.functional import cached_property
from django.views.generic.list import BaseListView

from ..components import ObjectsList, ChangeList
from ..utils.functools import just, call, first, pluck, \
    map_keys, map_values, reduce_dict, filter_dict
from ..utils.translation import _, get_string, get_model_names, verbose_choices
from .base import BaseView


class ListView(BaseView, BaseListView):
    components = {'content': ObjectsList}
    list_display = []

    def lookup_field(self, obj, field, next_obj=getattr, end_obj=getattr):
        if not '__' in field:
            return end_obj(obj, field)
        return self.lookup_field(
            next_obj(obj, field.split('__')[0]),
            '__'.join(field.split('__')[1:]),
            next_obj=next_obj,
            end_obj=end_obj)

    def get_model_field(self, field):
        try:
            return self.lookup_field(
                self.model,
                get_string(field),
                next_obj=lambda model, field: model._meta.get_field(field).rel.to,
                end_obj=lambda model, field: model._meta.get_field(field))
        except FieldDoesNotExist:
            return {'name': field}

    def get_list_view(self, op, value=None):
        return value or getattr(self, 'list_' + op)

    def lookup_field_display(self, obj, field, model=True):
        method = 'get_{0}_display'.format(field)
        if isinstance(obj, self.model) and hasattr(self, method):
            return getattr(self, method)(obj)
        elif model:
            return call(getattr(obj, method, lambda: getattr(obj, field)))
        raise AttributeError('Missing display data for "{0}"'.format(field))

    def display_column(self, field):
        if isinstance(field, dict):
            return field['name']
        elif isinstance(field, ManyToOneRel):
            return field.field.model._meta.verbose_name_plural
        else:
            return field.verbose_name

    def display_field(self, name, field, obj):
        if isinstance(field, dict):
            return self.lookup_field_display(
                obj,
                get_string(field['name']).replace(' ', '_'),
                model=False)
        elif '__' in get_string(name):
            return self.lookup_field(obj, name, end_obj=self.lookup_field_display)
        else:
            value = self.lookup_field_display(obj, field.name)
            if isinstance(field, ManyToOneRel) and not isinstance(value, str):
                return value.count()
            return value

    def get_thead_items(self):
        return list(map(
            self.display_column, map(
                self.get_model_field,
                self.get_list_view('display'))))

    def get_tbody_items(self, objects):
        return list(map(
            lambda obj: {
                'pk': obj.pk,
                'fields': list(map(
                    lambda name: self.display_field(
                        name, self.get_model_field(name), obj),
                    self.get_list_view('display')))},
            objects))

    def get_context_data(self, **kw):
        data = super().get_context_data(**kw)
        return dict(
            data,
            thead=self.get_thead_items(),
            tbody=self.get_tbody_items(data['object_list']))


class AdminView(ListView):
    components = {'content': ChangeList}
    list_filter = []
    list_counters = []
    list_sort = []
    list_search = []
    actions = [{
        'icon': 'plus-square',
        'title': _('Add'),
        'level': 'success',
        'link': {'rel': 'add'}}]

    @property
    def objects(self):
        if not hasattr(self, '_objects'):
            self._objects = self.get_queryset(ignore_state=True)
        return self._objects

    @property
    def filters(self):
        return self.get_filters()

    @property
    def filter_keys(self):
        return map(
            first, map(
                lambda name: self.get_filter(name, self.get_model_field(name)),
                self.get_list_view('filter')))

    @property
    def sort_keys(self):
        return map(
            lambda name: name.replace('__', '.'),
            self.get_list_view('sort'))

    @property
    def active_filters(self):
        return list(map(
            first, filter(
                lambda f: f[1] in self.request.GET,
                zip(
                    self.list_filter,
                    map(
                        lambda key: '-'.join(['filter', key]),
                        self.filter_keys)))))

    def get_state(self, method):
        return map_keys(
            lambda k: k.replace('.', '__'),
            map_values(
                lambda v: v if not v in ('True', 'False') else v[0] == 'T',
                filter_dict(
                    lambda k, v: k in list(getattr(self, '_'.join([method, 'keys']))),
                    map_keys(
                        lambda k: k.split('-')[1],
                        filter_dict(
                            lambda k, v: k.startswith(method),
                            self.request.GET)))))

    def get_queryset(self, ignore_state=False):
        qs = self.model.objects.all() if self.queryset is None else self.queryset
        if not ignore_state:
            states = {k: self.get_state(k) for k in ('filter', 'sort')}
            if states['filter']:
                qs = qs.filter(**states['filter'])
            if states['sort']:
                qs = qs.order_by(*reduce_dict(
                    lambda k, v: ('-' if v is False else '') + k,
                    states['sort']))
            if 'search' in self.request.GET:
                for word in self.request.GET['search'].split():
                    qs = qs.filter(reduce(
                        operator.or_, map(
                        lambda lookup: Q(**{lookup: word}),
                        map(
                            lambda f: '__'.join([f, 'icontains']),
                            self.get_list_view('search')))))
        return qs

    def get_list_view(self, op, value=None):
        if op == 'display':
            value = list(filter(
                lambda f: not f in self.active_filters,
                (value or self.list_display)))
        return super().get_list_view(op, value)

    def get_actions(self):
        return self.actions

    def get_generic_filter(self, name, choices):
        result = []
        name = name.replace('__', '.')
        values = pluck(self.objects, name)
        for choice in choices:
            count = values.count(choice[0])
            if count:
                result.append(list(chain(choice, [count])))
        return [name, result]

    def get_filter(self, name, field):
        storage = '_{0}_filter'.format(name)
        if not hasattr(self, storage):
            if isinstance(field, BooleanField):
                result = self.get_generic_filter(
                    name, [(True, _('Yes')), (False, _('No'))])
            elif field.is_relation:
                result = self.get_generic_filter(
                    name + '__pk', map(
                        lambda obj: (obj.pk, str(obj)),
                        field.rel.to.objects.all()))
            elif hasattr(field, 'choices'):
                result = self.get_generic_filter(name, field.choices)
            else:
                method = 'get_{0}_filter'.format(get_string(name))
                result = getattr(self, method)(name, field)
            setattr(self, storage, result)
        return getattr(self, storage)

    def get_filters(self):
        return list(map(
            lambda name: chain(
                [self.display_column(self.get_model_field(name))],
                self.get_filter(name, self.get_model_field(name)),
                [name in self.get_list_view('counters')]),
            self.get_list_view('filter')))

    def get_thead_items(self):
        return map(
            lambda name: [
                self.display_column(self.get_model_field(name)),
                name in self.get_list_view('sort') and name.replace('__', '.')],
            self.get_list_view('display'))

    def get_context_data(self, **kw):
        return dict(
            super().get_context_data(**kw),
            actions=self.get_actions(),
            filters=self.filters,
            search=bool(len(self.get_list_view('search'))))

    def get_redirect_features(self):
        if not self.objects:
            return self.navigate('-'.join([
                get_string(get_model_names(self.model)[0]),
                'add']))
        if self.list_filter and not self.request.GET:
            kw = {}
            for (title, key, choices, counters) in self.filters:
                choices = list(map(first, choices))
                if len(choices) == 1:
                    kw['-'.join(['filter', key])] = choices[0]
            if kw:
                return redirect('?'.join([self.request.path, urlencode(kw)]))

    def get(self, request, *args, **kwargs):
        return self.get_redirect_features() \
            or super().get(request, *args, **kwargs)
